from base64 import b64encode
from bitn import BitBin

class Splice_Info_Section:

    def __init__(self):
        pass

    def decode(self, bites):
        bitbin = BitBin(bites)
        self.table_id = bitbin.ashex(8)
        if self.table_id != '0xfc':
            raise ValueError('splice_info_section.table_id should be 0xfc')
        self.section_syntax_indicator = bitbin.asflag(1)
        self.private = bitbin.asflag(1)
        self.reserved = bitbin.ashex(2)
        if self.reserved != '0x3':
            raise ValueError('splice_info_section.reserved should be 0x3')
        self.section_length = bitbin.asint(12)
        self.protocol_version = bitbin.asint(8)
        if self.protocol_version != 0:
            raise ValueError('splice_info_section.protocol_version should be 0')
        self.encrypted_packet = bitbin.asflag(1)
        self.encryption_algorithm = bitbin.asint(6)
        self.pts_adjustment = bitbin.as90k(33)
        self.cw_index = bitbin.ashex(8)
        self.tier = bitbin.ashex(12)
        self.splice_command_length = bitbin.asint(12)
        self.splice_command_type = bitbin.asint(8)
        self.descriptor_loop_length = 0

    def encode(self):
        '''
        first byte is:
            table_id
        '''
        first_byte = int(self.table_id,16)
        bencoded = int.to_bytes(first_byte, 1, byteorder='big')
        '''
        two_bytes is:
            section_syntax_indicator
            private
            reserved
            section_length
        '''
        two_bytes = 0
        if self.section_syntax_indicator:
            two_bytes = (1 << 15)
        if self.private: two_bytes += (self.private << 14)
        two_bytes += (int(self.reserved,16) << 12)
        two_bytes += self.section_length
        bencoded += int.to_bytes(two_bytes, 2, byteorder='big')
        '''
        proto_byte is:
            protocol_version
        '''
        proto_byte = self.protocol_version
        bencoded += int.to_bytes(proto_byte, 1, byteorder='big')
        '''
        five_bytes is:
            encrypted_packet
            encryption_algorithm
            pts_adjustment
        '''
        five_bytes = 0
        if self.encrypted_packet:
            five_bytes = (1 << 39)
        five_bytes += (self.encryption_algorithm  << 33)
        five_bytes += int(self.pts_adjustment  * 90000)
        bencoded += int.to_bytes(five_bytes, 5, byteorder='big')
        '''
        cw_byte is:
            cw_index
        '''
        cw_byte = int(self.cw_index, 16)
        bencoded += int.to_bytes(cw_byte, 1, byteorder='big')
        '''
        three_bytes is:
            tier
            splice_command_length
        '''
        three_bytes = (int(self.tier, 16) << 12)
        three_bytes += self.splice_command_length
        bencoded += int.to_bytes(three_bytes, 3, byteorder='big')
        '''
        cmd_byte is:
            splice_command_type
        '''
        cmd_byte = self.splice_command_type
        bencoded += int.to_bytes(cmd_byte, 1, byteorder='big')
        print(bencoded)
        print(b64encode(bencoded))



from bitn import BitBin
from .splice import Splice
from .stream import Stream


class StreamPlus(Stream):
    '''
    StreamPlus adds PID and PTS for the SCTE 35 packets
    to the Stream class.
    '''
    NON_PTS_STREAM_IDS = [188, 190, 191, 240, 241, 242, 248]

    def __init__(self, tsdata, show_null = False):
        self.PTS = False
        super().__init__(tsdata,show_null)

    def parse_pts(self,bitbin):
        '''
        This is the process described in the official
        Mpeg-ts specification.
        '''
        a = bitbin.asint(3) << 30
        bitbin.forward(1)           
        b = bitbin.asint(15) << 15
        bitbin.forward(1)          
        c = bitbin.asint(15)
        d = (a+b+c)/90000.0
        # self.PTS is updated when we find a pts.
        self.PTS=round(d,6)
    
    def parse_pusi(self, packetdata):
        '''
        If the pusi data contains these markers,
        we can pull a PTS value..
        '''
        if packetdata[2] == 1: 
            if packetdata[3] not in self.NON_PTS_STREAM_IDS:  
                if (packetdata[6] >> 6) == 2: 
                    if (packetdata[7] >> 6) == 2:
                        if (packetdata[9] >> 4) == 2: 
                            bitbin = BitBin(packetdata[9:])
                            bitbin.forward(4)
                            self.parse_pts(bitbin)

    def parse_packet(self,packet):
        two_bytes = int.from_bytes(packet[1:3],byteorder='big')
        pid = hex(two_bytes & 0x1fff)
        if (two_bytes >> 14 & 0x1):
            self.parse_pusi(packet[4:20])
        if self.chk_magic(packet[:20]):
            packet_data = {'pid':pid,'pts':self.PTS}
            Splice(packet,packet_data).show()
    
    def parse(self,packets):
        [self.parse_packet(packet) for packet in packets]



from .decode import decode
#from .segmentation import Segmentation_Descriptor
from .splice import Splice
from .stream import Stream
from .streamplus import StreamPlus
from .splice_info_section import Splice_Info_Section

def i2b(i,wide):
    return int.to_bytes(i,wide,byteorder='big')



import sys
from .splice import Splice
from .stream import Stream


def read_stdin():
    scte35 = None
    try:
        scte35 = Stream(sys.stdin.buffer, show_null=False).decode()
    except BaseException:
        scte35 = Splice(sys.stdin.buffer)
        scte35.show()        
    return scte35

def read_stuff(stuff):
    scte35 = None
    try:
        scte35 = Splice(stuff)
        scte35.show()
    except BaseException:
        try:
            with open(stuff, 'rb') as tsdata:
                scte35 = Stream(tsdata,show_null=False).decode()
        except BaseException:
            pass        
    return scte35

def decode(stuff=None):
    """
    All purpose SCTE 35 decoder function
    
    usage:

    # for a mpegts video

    import threefive
    threefive.decode('/path/to/mpegts')

    # for a base64 encoded string

    import threefive
    Bee64='/DAvAAAAAAAA///wBQb+dGKQoAAZAhdDVUVJSAAAjn+fCAgAAAAALKChijUCAKnMZ1g='
    threefive.decode(Bee64)  
    """

    if stuff in [None, sys.stdin.buffer]:
        return read_stdin()
    else:
        return read_stuff(stuff)



from .splice import Splice
import sys

class Stream:
    '''
    Fast parse mpegts files and streams
    for SCTE 35 packets
    '''
    cmd_types = [4,5,6,7,255] # splice command types

    def __init__(self,tsdata, show_null = False):
        if show_null:
            self.cmd_types.append(0)
        self.tsdata = tsdata

    def decode(self):
        '''
        Split data into 188 byte packets
        '''
        pkt_sz = 188  # mpegts packet size
        pkt_ct = 256 # packet count
        chunk_sz = pkt_sz * pkt_ct
        while self.tsdata:
            chunk = self.tsdata.read(chunk_sz)
            if not chunk: break
            self.parse([chunk[i:i+pkt_sz]
                        for i in range(0,len(chunk),pkt_sz)])

    def chk_magic(self,pkt):
        '''
        Fast scte35 packet detection
        '''
        magic_bytes = b'\xfc0'
        if pkt[8] == 0: 
            if pkt[18] in self.cmd_types:
                return pkt[5:7] == magic_bytes
        return False
    
    def parse(self,packets):
        '''
        Parse scte35 packets
        '''
        [Splice(pkt).show() for pkt in filter(self.chk_magic,packets)]



from .descriptors import Splice_Descriptor
"""
Table 20 from page 58 of
https://www.scte.org/SCTEDocs/Standards/ANSI_SCTE%2035%202019r1.pdf
"""
table20 = {
    0x00: "Restrict Group 0",
    0x01: "Restrict Group 1",
    0x02: "Restrict Group 2",
    0x03: "No Restrictions",
}

"""
table 22 from page 62 of
https://www.scte.org/SCTEDocs/Standards/ANSI_SCTE%2035%202019r1.pdf
I am using the segmentation_type_id as a key.

Segmentation_type_id : segmentation_message
	
"""
table22 = {
    0x00: "Not Indicated",
    0x01: "Content Identification",
    0x10: "Program Start",
    0x11: "Program End",
    0x12: "Program Early Termination",
    0x13: "Program Breakaway",
    0x14: "Program Resumption",
    0x15: "Program Runover Planned",
    0x16: "Program RunoverUnplanned",
    0x17: "Program Overlap Start",
    0x18: "Program Blackout Override",
    0x19: "Program Start ??? In Progress", 
    0x20: "Chapter Start",
    0x21: "Chapter End",
    0x22: "Break Start",
    0x23: "Break End", 
    0x24: "Opening Credit Start",
    0x25: "Opening Credit End", 
    0x26: "Closing Credit Start",
    0x27: "Closing Credit End",
    0x30: "Provider Advertisement Start",
    0x31: "Provider Advertisement End",
    0x32: "Distributor Advertisement Start",
    0x33: "Distributor Advertisement End", 
    0x34: "Provider Placement Opportunity Start",
    0x35: "Provider Placement Opportunity End", 
    0x36: "Distributor Placement Opportunity Start",
    0x37: "Distributor Placement Opportunity End", 
    0x38: "Provider Overlay Placement Opportunity Start",
    0x39: "Provider Overlay Placement Opportunity End",
    0x3A: "Distributor Overlay Placement Opportunity Start",
    0x3B: "Distributor Overlay Placement Opportunity End",
    0x40: "Unscheduled Event Start", 
    0x41: "Unscheduled Event End", 
    0x50: "Network Start", 
    0x51: "Network End", 
    0x3B: "Distributor Overlay Placement Opportunity End",
    0x40: "Unscheduled Event Start", 
    0x41: "Unscheduled Event End", 
    0x50: "Network Start",
    0x51: "Network End"}

       
class Segmentation_Descriptor(Splice_Descriptor):
    """
    Table 19 - segmentation_descriptor()
    """
    def __init__(self, bitbin, tag):
        if not super().__init__(bitbin, tag): return False
        bitbin = bitbin
        self.name = "Segmentation Descriptor"
        self.segmentation_event_id = bitbin.ashex(32) # 4 bytes
        self.segmentation_event_cancel_indicator = bitbin.asflag(1)
        bitbin.forward(7) # 1 byte
        if not self.segmentation_event_cancel_indicator:
            self.set_flags(bitbin) # 1 byte
            if not self.program_segmentation_flag:
                self.component_count = bitbin.asint(8)# 1 byte
                self.components = []
                for i in range(0, self.component_count): # 6 bytes each
                    comp = {}
                    comp["component_tag"] = bitbin.asint(8)
                    bitbin.forward(7)
                    comp["pts_offset"] = bitbin.as90k(33)
                    self.components.append(comp)
            self.set_segmentation(bitbin)  
      
    def set_flags(self,bitbin):  # 1 byte for set flags
        self.program_segmentation_flag = bitbin.asflag(1)
        self.segmentation_duration_flag = bitbin.asflag(1)
        self.delivery_not_restricted_flag = bitbin.asflag(1)
        if not self.delivery_not_restricted_flag:
            self.web_delivery_allowed_flag = bitbin.asflag(1)
            self.no_regional_blackout_flag = bitbin.asflag(1)
            self.archive_allowed_flag = bitbin.asflag(1)
            self.device_restrictions = table20[bitbin.asint(2)]
        else:
            bitbin.forward(5)

    def set_segmentation(self,bitbin):
        if self.segmentation_duration_flag:
            self.segmentation_duration = bitbin.as90k(40)# 5 bytes

        self.segmentation_upid_type = bitbin.asint(8) # 1 byte

        self.segmentation_upid_length = bitbin.asint(8) # 1 byte
        
        self.segmentation_upid = self.set_segmentation_upid(bitbin, 
            self.segmentation_upid_type, self.segmentation_upid_length)
        
        self.segmentation_type_id = bitbin.asint(8) # 1 byte

        if self.segmentation_type_id in table22.keys():
            self.segmentation_message = table22[self.segmentation_type_id]
            self.set_segments(bitbin)
        bitbin = None   
            
    def set_segmentation_upid(self,bitbin,upid_type,upid_length):
        upid_map={
            0x02: ['Deprecated', self.AdID],
            0x03: ['Ad ID',self.AdID],
            0x04: ['UMID', self.UMID],
            0x05: ['ISAN',self.ISAN],
            0x06: ['ISAN',self.ISAN],
            0x07: ['TID',self.TID],
            0x08: ['AiringID',self.AirID],
            0x09: ['ADI',self.ADI],
            0x0a: ['EIDR',self.EIDR],
            0x0b: ['ATSC',self.ATSC],
            0x0c: ['MPU',self.MPU],
            0x0d: ['MID',self.MID],
            0x0e: ['ADS Info',self.ADS],
            0x0f: ['URI',self.URI]
            }
        upid_id=""
        if upid_type in upid_map.keys():
            upid_id= upid_map[upid_type][1](bitbin,upid_length)
            if upid_type != 0x09 :
                return f'{upid_map[upid_type][0]}:{upid_id}'
        return upid_id

    def set_segments(self,bitbin): 
        self.segment_num = bitbin.asint(8) # 1 byte
        self.segments_expected = bitbin.asint(8) # 1 byte
        if self.segmentation_type_id in [0x34, 0x36, 0x38, 0x3a]:
            '''
            if there are 16 more bits in bitbin, read them.
            '''
            if bitbin.idx >= 16:
                self.sub_segment_num = bitbin.asint(8) # 1 byte
                self.sub_segments_expected = bitbin.asint(8)# 1 byte
            else:
                self.sub_segment_num = self.sub_segments_expected = 0
                
    def ADI(self,bitbin,upid_length):
        return self.URI(bitbin,upid_length)
        
    def AdID(self, bitbin, upid_length):
        return self.URI(bitbin,upid_length)

    def ADS(self, bitbin, upid_length):
        return self.URI(bitbin,upid_length)
    
    def AirID(self, bitbin, upid_length):
        return bitbin.ashex(upid_length << 3)

    def ATSC(self, bitbin, upid_length):
         return {'TSID': bitbin.asint(16),
                'reserved': bitbin.asint(2),
                'end_of_day':bitbin.asint(5),
                'unique_for':bitbin.asint(9),
                'content_id': bitbin.asdecodedhex((upid_length -4) << 3)}
      
    def ISAN(self, bitbin, upid_length):
        pre = '0000-0000-'
        middle = bitbin.ashex(upid_length << 3)
        post = '-0000-Z-0000-0000-6'
        return f'{pre}{middle[2:6]}{post}'
        
    def MID(self, bitbin, upid_length):
        upids = []
        bitcount = (upid_length << 3)
        while bitcount > 0:
            upid_type = bitbin.asint(8) # 1 byte
            bitcount -= 8
            upid_length = bitbin.asint(8)
            bitcount -= 8
            segmentation_upid = self.set_segmentation_upid(bitbin,upid_type,upid_length)
            bitcount -= (upid_length << 3)
            upids.append(segmentation_upid)
        return upids     
        
    def MPU(self, bitbin, upid_length):
        bitcount = (upid_length << 3 )
        return {'format identifier':bitbin.asint(32), 
                'private data':bitbin.asint(bitcount -32)}

    def EIDR(self, bitbin, upid_length):
        pre = bitbin.asint(16)
        post = bitbin.ashex(80)
        return f'10.{pre}/{post[2:6]}-{post[6:10]}-{post[10:14]}-{post[14:18]}-T'        

    def TID(self, bitbin, upid_length):
        return self.URI(bitbin, upid_length)

    def UMID(self, bitbin, upid_length):
        n = 8
        pre = ''.join(self.AirID(upid_length).split('x',1))
        return '.'.join([pre[i:i+n] for i in range(0, len(pre), n)])

    def URI(self, bitbin, upid_length):
        if upid_length > 0: return bitbin.asdecodedhex(upid_length << 3)
        return None



class Splice_Descriptor:
    def __init__(self,bitbin,tag):
        self.tag = tag
        # identiﬁer 32 uimsbf == 0x43554549 (ASCII “CUEI”)
        self.identifier = bitbin.asdecodedhex(32)
        if self.identifier != "CUEI":
            raise ValueError(
                'All descriptors must have a identifier of "CUEI"')
        else:
            return self.identifier


class Avail_Descriptor(Splice_Descriptor):
    """
    Table 17 -  avail_descriptor()
    """
    def __init__(self, bitbin, tag):
        if not super().__init__(bitbin, tag):
            return False
        self.name = "Avail Descriptor"
        self.provider_avail_id = bitbin.asint(32)


class Dtmf_Descriptor(Splice_Descriptor):
    """
    Table 18 -  DTMF_descriptor()
    """
    def __init__(self, bitbin, tag):
        if not super().__init__(bitbin, tag):
            return False
        self.name = "DTMF Descriptor"
        self.preroll = bitbin.asint(8)
        self.dtmf_count = bitbin.asint(3)
        bitbin.forward(5)
        self.dtmf_chars = []
        for i in range(0, self.dtmf_count):
            self.dtmf_chars.append(bitbin.asint(8))

     
class Time_Descriptor(Splice_Descriptor):
    """
    Table 25 - time_descriptor()
    """
    def __init__(self, bitbin, tag):
        if not super().__init__(bitbin, tag):
            return False
        self.name = "Time Descriptor"
        self.TAI_seconds = bitbin.asint(48)
        self.TAI_ns = bitbin.asint(32)
        self.UTC_offset = bitbin.asint(16)


class Audio_Descriptor(Splice_Descriptor):
    """
    Table 26 - audio_descriptor()
    """
    def __init__(self, bitbin, tag):
        if not super().__init__(bitbin, tag):
            return False
        self.name = "Audio Descriptor"
        self.components = []
        self.audio_count = bitbin.asint(4)
        reserved = bitbin.asint(4)
        for i in range(0, self.audio_count):
            comp = {}
            comp["component_tag"] = bitbin.asint(8)
            comp["ISO_code="] = bitbin.asint(24)
            comp["bit_stream_mode"] = bitbin.asint(3)
            comp["num_channels"] = bitbin.asint(4)
            comp["full_srvc_audio"] = bitbin.asflag(1)
            self.components.append(comp)



from base64 import b64decode
from bitn import BitBin
import json
from threefive.segmentation import Segmentation_Descriptor
from threefive.splice_info_section import Splice_Info_Section
from threefive.descriptors import (
    Avail_Descriptor,
    Dtmf_Descriptor,
    Time_Descriptor,
    Audio_Descriptor
    )
from threefive.splice_commands import (
    Splice_Null,
    Splice_Schedule,
    Splice_Insert,
    Time_Signal,
    Bandwidth_Reservation,
    Private_Command 
    )


class Splice:
    '''
    The threefive.Splice class handles parsing
    SCTE 35 message strings.
    '''
    # map of known descriptors and associated classes
    descriptor_map = { 0: Avail_Descriptor,
                       1: Dtmf_Descriptor,
                       2: Segmentation_Descriptor,
                       3: Time_Descriptor,
                       4: Audio_Descriptor }

    # map of known splice commands and associated classes
    command_map = { 0: Splice_Null,
                    4: Splice_Schedule,
                    5: Splice_Insert,
                    6: Time_Signal,
                    7: Bandwidth_Reservation,
                  255: Private_Command }

    def __init__(self, data, packet_data=None):
        payload = self.mk_payload(data)
        self.packet_data = packet_data
        self.decode(payload)
 
    def decode(self,payload):
        payload = self.mk_info_section(payload)
        payload = self.mk_command(payload)
        payload = self.mk_descriptors(payload)
        self.info_section.crc = hex(int.from_bytes(payload[0:4],
                                             byteorder = 'big'))
        
    def mk_payload(self,data):
        if data[0] == 0x47:
            payload = data[5:]
        else:
            payload = self.mkbits(data)  
        return payload

    def mk_info_section(self,payload):
        info_size = 14
        info_payload = payload[:info_size]
        self.info_section = Splice_Info_Section()
        self.info_section.decode(info_payload)
        return payload[info_size:]

    def mk_command(self,payload):
        cmdl = self.info_section.splice_command_length
        # fix for bad self.info_section.splice_command_length
        cmdbb = BitBin(payload)
        self.set_splice_command(cmdbb)
        cmdl = self.info_section.splice_command_length = self.command.splice_command_length
        return payload[cmdl:]

    def mk_descriptors(self,payload):
        self.descriptors = []
        dll = self.info_section.descriptor_loop_length = int.from_bytes(payload[0:2],byteorder = 'big')
        payload = payload[2:]
        self.descriptorloop(payload,dll)
        return payload[dll:]
        
    def __repr__(self):
        return str(self.get())

    def descriptorloop(self,payload,dll):
        '''
        parses all splice descriptors
        '''
        while dll > 0:
            try:
                sd = self.set_splice_descriptor(payload)
                sdl = sd.descriptor_length
                dll-= sdl+2
                payload = payload[sdl+2:]
                self.descriptors.append(sd)
            except:
                dll = -1

    def get(self):
        '''
        Returns a dict of dicts for all three parts
        of a SCTE 35 message.
        '''
        scte35 = {  'info_section':self.get_info_section(),
                    'command':self.get_command(),
                    'descriptors':self.get_descriptors()}

        if self.packet_data:
            scte35.update(self.get_packet_data())
        return scte35

    def get_command(self):
        '''
        returns the SCTE 35
        splice command data as a dict.
        '''
        return self.kvclean(vars(self.command))

    def get_descriptors(self):
        '''
        Returns a list of SCTE 35
        splice descriptors as dicts.
        '''
        return [self.kvclean(vars(d)) for d in self.descriptors]
    
    def get_info_section(self):
        '''
        Returns SCTE 35
        splice info section as a dict
        '''
        return self.kvclean(vars(self.info_section))

    def get_packet_data(self):
        return self.kvclean(self.packet_data)

    def kvclean(self,obj):
        '''
        kvclean removes items from a dict if the value is None
        '''
        return {k: v for k, v in obj.items() if v is not None}

    def kvprint(self, obj):
        print('\n')
        print(json.dumps(obj,indent = 2))

    def mkbits(self, s):
        '''
        Convert Hex and Base64 strings into bytes.
        '''
        if s[:2].lower() == '0x':
            s = s[2:]
        if s[:2].lower() == 'fc':
            return bytes.fromhex(s)
        try:
            return b64decode(s)
        except:
            return s

    def set_splice_command(self,cmdbb):
        '''
        Splice Commands looked up in self.command_map
        '''
        sct = self.info_section.splice_command_type
        if sct not in self.command_map.keys():
            raise ValueError('Unknown Splice Command Type')
            return False
        self.command = self.command_map[sct]()
        self.command.decode(cmdbb)

    def set_splice_descriptor(self,payload):
        '''
        Splice Descriptors looked up in self.descriptor_map
        '''
        # splice_descriptor_tag 8 uimsbf
        tag = payload[0]
        desc_len = payload[1]
        payload = payload[2:]
        bitbin = BitBin(payload[:desc_len])
        payload = payload[desc_len:]
        if tag in self.descriptor_map.keys():
            sd = self.descriptor_map[tag](bitbin,tag)
            sd.descriptor_length = desc_len
            return sd
        else:
            return False

    def show(self):
        '''
        pretty prints the SCTE 35 message
        '''
        self.kvprint(self.get())

    def show_command(self):
        '''
        pretty prints SCTE 35 splice command
        '''
        self.kvprint(self.get_command())

    def show_descriptors(self):
        '''
        pretty prints SCTE 35 splice descriptors
        '''
        self.kvprint(self.get_descriptors())

    def show_info_section(self):
        '''
        pretty prints SCTE 35 splice info section
        '''
        self.kvprint(self.get_info_section())



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="threefive",
    version="2.1.49",
    author="fu-corp",
    author_email="spam@futzu.com",
    description="A Fast SCTE 35 Decoder for Mpeg-TS Video, and Base64 or Hex Encoded Messages.", 
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/futzu/threefive",
    packages=setuptools.find_packages(),
    install_requires=["bitn>=0.0.27",],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



class Splice_Command:
    '''
    Base class for all splice command classes,
    not used directly.
    '''
    def __init__(self):
        pass

    def decode(self, bitn):
        pass

    def parse_break(self, bitbin):
        self.break_auto_return = bitbin.asflag(1)
        bitbin.forward(6)
        self.break_duration = bitbin.as90k(33)

    def encode_break(self): #40bits
        break_bytes=0
        if self.break_auto_return:
            break_bytes = 1 << 39
        break_bytes += (self.break_duration * 90000)
        return int.to_bytes(break_bytes, 5, byteorder='big')

    def splice_time(self, bitbin):  # 40bits
        self.time_specified_flag = bitbin.asflag(1)
        if self.time_specified_flag:
            bitbin.forward(6)
            self.pts_time = bitbin.as90k(33)
        else:
            bitbin.forward(7)

    def encode_splice_time(self):
        st_bytes=0
        if self.time_specified_flag:
            st_bytes = 1 << 39
            st_bytes += (self.pts_time * 90000)
        return int.to_bytes(st_bytes, 5, byteorder='big')


class Splice_Null(Splice_Command):
    """
    Table 7 - splice_null()
    """
    def decode(self, bitbin):
        self.name = "Splice Null"
        self.splice_command_length = 0


class Splice_Schedule(Splice_Command):
    """
    Table 8 - splice_schedule()
    """
    def decode(self, bitbin):
        bit_start = bitbin.idx
        self.name = "Splice Schedule"
        splice_count = bitbin.asint(8)
        for i in range(0, splice_count):
            self.splice_event_id = bitbin.asint(32)
            self.splice_event_cancel_indicator = bitbin.asflag(1)
            bitbin.forward(7)
            if not self.splice_event_cancel_indicator:
                self.out_of_network_indicator = bitbin.asflag(1)
                self.program_splice_flag = bitbin.asflag(1)
                self.duration_flag = bitbin.asflag(1)
                bitbin.forward(5)
                if self.program_splice_flag:
                    self.utc_splice_time = bitbin.asint(32)
                else:
                    self.component_count = bitbin.asint(8)
                    self.components = []
                    for j in range(0, self.component_count):
                        self.components[j] = {
                            "component_tag": bitbin.asint(8),
                            "utc_splice_time": bitbin.asint(32),
                        }
                if self.duration_flag: self.break_duration(bitbin)
                self.unique_program_id = bitbin.asint(16)
                self.avail_num = bitbin.asint(8)
                self.avails_expected = bitbin.asint(8)
        bit_end = bitbin.idx
        self.splice_command_length = int((bit_start - bit_end) >>3)


class Splice_Insert(Splice_Command):
    """
    Table 9 - splice_insert()
    """
    def __init__(self):
        self.splice_event_id = None
        self.splice_event_cancel_indicator = None
        self.out_of_network_indicator = None
        self.program_splice_flag = None
        self.duration_flag = None
        self.splice_immediate_flag = None
        self.component_count = 0
        self.components = []
        self.time_specified_flag = None
        self.pts_time = None
        self.break_auto_return = None
        self.break_duration = None
        self.unique_program_id = None
        self.avail_num = None
        self.avail_expected = None

    def decode(self, bitbin):
        bit_start = bitbin.idx
        self.name = "Splice Insert"
        self.splice_event_id = bitbin.asint(32) # uint32
        self.splice_event_cancel_indicator = bitbin.asflag(1)
        bitbin.forward(7) #uint8
        if not self.splice_event_cancel_indicator:
            self.out_of_network_indicator = bitbin.asflag(1)
            self.program_splice_flag = bitbin.asflag(1)
            self.duration_flag = bitbin.asflag(1)
            self.splice_immediate_flag = bitbin.asflag(1)
            bitbin.forward(4) #uint8
            if self.program_splice_flag and not self.splice_immediate_flag:
                self.splice_time(bitbin) # uint8 + uint32
            if not self.program_splice_flag:
                self.component_count = bitbin.asint(8)# uint 8
                self.components = []
                for i in range(0, self.component_count):
                    self.components[i] = bitbin.asint(8)
                if not self.splice_immediate_flag: self.splice_time(bitbin)
            if self.duration_flag: self.parse_break(bitbin)
            self.unique_program_id = bitbin.asint(16)
            self.avail_num = bitbin.asint(8)
            self.avail_expected = bitbin.asint(8)
            bit_end = bitbin.idx
            self.splice_command_length = int((bit_start - bit_end) >>3)
            

class Time_Signal(Splice_Command):
    """
    Table 10 - time_signal()
    """
    def __init__(self):
        self.time_specified_flag = None
        self.pts_time = None

    def decode(self, bitbin):
        bit_start = bitbin.idx
        self.name = "Time Signal"
        self.splice_time(bitbin)
        bit_end = bitbin.idx
        self.splice_command_length = int((bit_start - bit_end) >>3)

    def encode(self):
        command_bytes =self.encode_splice_time()


class Bandwidth_Reservation(Splice_Command):
    """
    Table 11 - bandwidth_reservation()
    """
    def decode(self, bitbin):
        bit_start = bitbin.idx
        self.name = "Bandwidth Reservation"
        bit_end = bitbin.idx
        self.splice_command_length = int((bit_start - bit_end) >>3)


class Private_Command(Splice_Command):
    """
    Table 12 - private_command()
    """
    def __init__(self):
        self.identifier = None

    def decode(self, bitbin):
        bit_start = bitbin.idx
        self.name = "Private Command"
        self.identifier = bitbin.asint(32)
        bit_end = bitbin.idx
        self.splice_command_length = int((bit_start - bit_end) >>3)

    def encode(self):
        command_bytes = int.to_bytes(self.identifier, 4, byteorder='big')
